/*  
 * Copyright (c) 2005 Blueprint Software Works
 *                    http://www.blueprintsoftwareworks.com
 * Copyright (c) 2005 Vitali E. Pelenyov
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
 * THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
 */

#include <string.h>
#include <malloc.h>

#include "debug.hpp"

#include "ndp.hpp"

#include "resource.hpp"

const NDPROPERTYINFO Properties[] =
{
    {ND_PROP_STRING, 0, "URL", ""},
    {ND_PROP_STRING, 0, "USERNAME", ""},
    {ND_PROP_STRING, 0, "PASSWORD", ""},
    {ND_PROP_STRING, 0, NULL, NULL},
};

const char *NdpTypes[] =
{
    "DAV",
    NULL
};

const NDPROPERTYINFO *NdpPropertiesInfo[] =
{
    Properties
};

const unsigned long NdpAttribute = NDPA_REUSE_FILE_CONNECTIONS;

PLUGINHELPERTABLE2 *ph = NULL;
int fEnableLog = 1;

int APIENTRY NdpPluginLoad (PLUGINHELPERTABLE2 *pPHT2)
{
    ph = pPHT2;
    
    if (ph->cb < sizeof (PLUGINHELPERTABLE2))
    {
        return ERROR_INVALID_FUNCTION;
    }
    
    return ResourceInit ();
}

int APIENTRY NdpPluginFree (void)
{
    return ResourceShutdown ();
}

int APIENTRY NdpMountResource (HRESOURCE *presource, int type, NDPROPERTYHANDLE properties)
{
    int rc = NO_ERROR;

    // 'type' is valid here and we support only one type so do not check it.
    (void)type;
    
    NdResource *p = new Resource (properties);
    
    if (p == NULL)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = p->Initialize ();
        
        if (rc == NO_ERROR)
        {
            *presource = (HRESOURCE)p;
        }
        else
        {
            delete p;
        }
    }
    
    LOG((ND_LL_DEBUG, "NdpMountResource: rc=%d, res = %p\r\n", rc, p));

    return rc;
}


int APIENTRY NdpFreeResource (HRESOURCE resource)
{
    LOG((ND_LL_DEBUG, "NdpFreeResource: res = %p\r\n", resource));
    
    delete (NdResource *)resource;

    LOG((ND_LL_DEBUG, "NdpFreeResource: res = %p deleted\r\n", resource));
    return NO_ERROR;
}


int APIENTRY NdpCreateConnection (HRESOURCE resource, HCONNECTION *pconn)
{
    int rc = NO_ERROR;

    NdConnection *p = new Connection ((Resource *)resource);
    
    if (p == NULL)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = p->Initialize ();
        
        if (rc == NO_ERROR)
        {
            *pconn = (HCONNECTION)p;
        }
        else
        {
            delete p;
        }
    }
    
    LOG((ND_LL_DEBUG, "NdpCreateConnection: rc = %d, conn = %p, resource = %p\r\n", rc, p, resource));

    return rc;
}

int APIENTRY NdpFreeConnection (HCONNECTION conn)
{
    LOG((ND_LL_DEBUG, "NdpFreeConnection: conn = %p\r\n", conn));
    
    delete (NdConnection *)conn;
    
    return NO_ERROR;
}

int APIENTRY NdpRsrcCompare (HRESOURCE resource, HRESOURCE resource2)
{
    NdResource *pRes = (NdResource *)resource;
    NdResource *pRes2 = (NdResource *)resource2;
    
    if (pRes->EqualTo (pRes2))
    {
        return ND_RSRC_EQUAL;
    }
    
    return ND_RSRC_DIFFERENT;
}

int APIENTRY NdpRsrcUpdate (HRESOURCE resource, HRESOURCE resource2)
{
    NdResource *pRes = (NdResource *)resource;
    NdResource *pRes2 = (NdResource *)resource2;
    
    // do nothing
    
    return NO_ERROR;
}

int APIENTRY NdpRsrcQueryInfo (HRESOURCE resource, ULONG *pulFlags, void *pdata, ULONG insize, ULONG *poutlen)
{
    NdResource *pRes = (NdResource *)resource;
    
    if (pRes->Available ())
    {
        *pulFlags = ND_RSRC_AVAILABLE;
    }
    
    char *info;
    
    int rc = pRes->QueryInfoString (&info);
    
    LOG((ND_LL_DEBUG, "NdpRsrcQueryInfo: rc = %d, info = %s\r\n", rc, info));
    
    if (rc == NO_ERROR)
    {
        *poutlen = strlen (info) + 1;
        
        if (*poutlen > insize)
        {
            rc = ERROR_BUFFER_OVERFLOW;
        }
        else
        {
            memcpy (pdata, info, *poutlen);
        }
            
        pRes->ReleaseInfoString (info);
    }
    
    return rc;
}

int APIENTRY NdpRsrcQueryFSAttach (HRESOURCE resource, void *pdata, ULONG insize, ULONG *poutlen)
{
    ULONG ulDummy = 0;
    return NdpRsrcQueryInfo (resource, &ulDummy, pdata, insize, poutlen);
}

int APIENTRY NdpRsrcQueryFSAllocate (HRESOURCE resource, NDFSALLOCATE *pfsa)
{
    (void)resource;
    (void)pfsa;
    return ERROR_NOT_SUPPORTED;
}

int APIENTRY NdpRefresh (HCONNECTION conn, char *path, int tree)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (path);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        pConn->Refresh (fullpath, tree);
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
    
    return NO_ERROR;
}

int APIENTRY NdpQueryPathInfo (HCONNECTION conn, void *plist, char *szPath)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    NdPathInfo pathinfo;
    
    NDPATHELEMENT *pel = ph->fsphNameElem (0);
        
    /* construct parent path, that will be needed later */
    char *parentPath = ""; // assume there is no parent path
    int l = 0;
        
    if (!ph->fsphIsLastElem (0, NULL))
    {
        /* this is not only element in the path */
            
        /* "aa\bb" -> "aa" */
        l = strlen (szPath) - pel->length - 1;
            
        parentPath = (char *)alloca (l + 1 + 1); // reserve space for a '\'
                
        memcpy (parentPath, szPath, l);
               
        parentPath[l] = '\0';
    }
                    
    /* check wildcards in the last (name) element of the path */
    int wildCards = (ph->fsphStrChr (pel->name, '*') != NULL) || 
                    (ph->fsphStrChr (pel->name, '?') != NULL);
                        
    if (wildCards)
    {
        /* wildcards specified, we must check only parent path. 
         * If the parent path exists and it is a directory, we return
         * ERROR_FILE_NOT_FOUND otherwise ERROR_PATH_NOT_FOUND
         */
        if (parentPath[0] == '\0')
        {
            /* parent is a mount point, which is a directory and exists */
            rc = ERROR_FILE_NOT_FOUND;
        }
        else
        {
            /* query parent info only for non-root paths */
            const char *fullpath = pConn->GetResource()->QueryFullPath (parentPath);
     
            if (!fullpath)
            {
                rc = ERROR_NOT_ENOUGH_MEMORY;
            }
            else
            {
                rc = pConn->QueryPathInfo (&pathinfo, fullpath);
                pConn->GetResource()->ReleaseFullPath (fullpath);
            }
               
            if (rc == NO_ERROR)
            {
                if (pathinfo.isDir ())
                {
                    rc = ERROR_FILE_NOT_FOUND;
                }
                else
                {
                    rc = ERROR_PATH_NOT_FOUND;
                }
            }
            else if (rc != ERROR_REM_NOT_LIST)
            {
                rc = ERROR_PATH_NOT_FOUND;
            }
        }
    }
    else
    {
        // no wildcards, a file or a directory
        const char *fullpath = pConn->GetResource()->QueryFullPath (szPath);
     
        if (!fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pConn->QueryPathInfo (&pathinfo, fullpath);

            pConn->GetResource()->ReleaseFullPath (fullpath);
        }
               
        if (rc == NO_ERROR)
        {
            /* path exists, so add its information to the plist */
            FILESTATUS3L stat;
            
            pathinfo.toFILESTATUS3L (&stat);
            
            ph->fsphAddFile32L (plist, &stat, pel->name, pel->length, NULL, 0, 0);
        }
        else if (rc != ERROR_REM_NOT_LIST)
        {
            if (parentPath[0] == '\0')
            {
                rc = ERROR_FILE_NOT_FOUND;
            }
            else
            {
                /* there is parent path, lets check it */
                const char *fullpath = pConn->GetResource()->QueryFullPath (parentPath);
     
                if (!fullpath)
                {
                    rc = ERROR_NOT_ENOUGH_MEMORY;
                }
                else
                {
                    rc = pConn->QueryPathInfo (&pathinfo, fullpath);
                        
                    pConn->GetResource()->ReleaseFullPath (fullpath);
                }
               
                if (rc == NO_ERROR)
                {
                    if (pathinfo.isDir ())
                    {
                        rc = ERROR_FILE_NOT_FOUND;
                    }
                    else
                    {
                        rc = ERROR_PATH_NOT_FOUND;
                    }
                }
                else if (rc != ERROR_REM_NOT_LIST)
                {
                    rc = ERROR_PATH_NOT_FOUND;
                }
            }
        }
    }
        
    LOG((ND_LL_DEBUG, "NdpQueryPathInfo: rc = %d, conn=%p, plist=%p, path = %s\r\n", rc, conn, plist, szPath));
    
    return rc;
}

int APIENTRY NdpDeletePathInfo (HRESOURCE resource, NDFILEINFO *pfi)
{
    (void)resource;
    (void)pfi;
    return NO_ERROR;
}

int APIENTRY NdpDiscardResourceData (HRESOURCE resource, NDDATABUF *pdatabuf)
{
    (void)resource;
    (void)pdatabuf;
    return NO_ERROR;
}

int APIENTRY NdpSetPathInfo (HCONNECTION conn, NDFILEINFO *pfi, char *szPathName)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    NDPATHELEMENT *pel = ph->fsphNameElem (0);
    
    NdPathInfo pathinfo(&pfi->stat, pel->name, pel->length);
        
    int rc = NO_ERROR;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (szPathName);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = pConn->SetPathInfo (fullpath, &pathinfo);
    
        LOG((ND_LL_DEBUG, "NdpSetPathInfo: rc = %d, conn = %p, path = %s\r\n", rc, conn, szPathName));
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
    
    return rc;
}

int APIENTRY NdpEAQuery (HCONNECTION conn, GEALIST *pGEAList, NDFILEINFO *pfi, FEALIST *pFEAList)
{
    (void)conn;
    (void)pGEAList;
    (void)pfi;
    (void)pFEAList;
    return ERROR_NOT_SUPPORTED;
}

int APIENTRY NdpEASet (HCONNECTION conn, FEALIST *pFEAList, NDFILEINFO *pfi)
{
    (void)conn;
    (void)pFEAList;
    (void)pfi;
    return ERROR_NOT_SUPPORTED;
}

int APIENTRY NdpEASize (HCONNECTION conn, NDFILEINFO *pfi, ULONG *pulEASize)
{
    (void)conn;
    (void)pfi;
    (void)pulEASize;
    return ERROR_NOT_SUPPORTED;
}

int APIENTRY NdpSetCurrentDir (HCONNECTION conn, NDFILEINFO *pfi, char *szPath)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfi;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (szPath);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = pConn->SetCurrentDir (fullpath);
    
        LOG((ND_LL_DEBUG, "NdpSetCurrentDir: rc = %d, conn=%p, pfi=%p, path = %s\r\n", rc, conn, pfi, szPath));
    
        if (rc != NO_ERROR)
        {
            rc = ERROR_ACCESS_DENIED;
        }
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
    
    return rc;
}

int APIENTRY NdpCopy (HCONNECTION conn, NDFILEINFO *pfiDst, char *szDst, NDFILEINFO *pfiSrc, char *szSrc, ULONG ulOption)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiDst;
    (void)pfiSrc;
    
    const char *src_fullpath = pConn->GetResource ()->QueryFullPath (szSrc);
    
    if (!src_fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        const char *dst_fullpath = pConn->GetResource ()->QueryFullPath (szDst);
    
        if (!dst_fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pConn->Copy (dst_fullpath, src_fullpath, ulOption);

            LOG((ND_LL_DEBUG, "NdpCopy: rc = %d, conn=%p, pfiDst = %p, dst = %s, pfiSrc = %p, src = %s, option = %8.8X\r\n", rc, conn, pfiDst, szDst, pfiSrc, szSrc, ulOption));

            if (rc != NO_ERROR)
            {
                rc = ERROR_CANNOT_COPY;
            }
            
            pConn->GetResource ()->ReleaseFullPath (dst_fullpath);
        }
        
        pConn->GetResource ()->ReleaseFullPath (src_fullpath);
    }
    
    return rc;
}

int APIENTRY NdpCopy2 (HCONNECTION conn, HRESOURCE resDst, NDFILEINFO *pfiDst, char *szDst, NDFILEINFO *pfiSrc, char *szSrc, ULONG ulOption)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiDst;
    (void)pfiSrc;
    
    const char *src_fullpath = pConn->GetResource ()->QueryFullPath (szSrc);
    
    if (!src_fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        const char *dst_fullpath = pConn->GetResource ()->QueryFullPath (szDst);
    
        if (!dst_fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pConn->CopyTo ((NdResource *)resDst, dst_fullpath, src_fullpath, ulOption);

            LOG((ND_LL_DEBUG, "NdpCopy2: rc = %d, conn=%p, resDst = %p, pfiDst = %p, dst = %s, pfiSrc = %p, src = %s, option = %8.8X\r\n", rc, conn, resDst, pfiDst, szDst, pfiSrc, szSrc, ulOption));
    
            if (rc != NO_ERROR)
            {
                rc = ERROR_CANNOT_COPY;
            }
            
            pConn->GetResource ()->ReleaseFullPath (dst_fullpath);
        }
        
        pConn->GetResource ()->ReleaseFullPath (src_fullpath);
    }
    
    return rc;
}

int APIENTRY NdpForceDelete (HCONNECTION conn, NDFILEINFO *pfi, char *szFile)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfi;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (szFile);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = pConn->DeleteFile (fullpath);
    
        LOG((ND_LL_DEBUG, "NdpForceDelete: rc = %d, conn=%p, pfi=%p, path = %s\r\n", rc, conn, pfi, szFile));
    
        if (rc != NO_ERROR)
        {
            rc = ERROR_ACCESS_DENIED;
        }
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
    
    return rc;
}

int APIENTRY NdpCreateDir (HCONNECTION conn, NDFILEINFO *pfiparent, char *szDirName, FEALIST *pFEAList)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiparent;
    (void)pFEAList;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (szDirName);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = pConn->CreateDir (fullpath);
    
        LOG((ND_LL_DEBUG, "NdpCreateDir: rc = %d, conn=%p, pfiparent=%p, path = %s, pFEAList = %p\r\n", rc, conn, pfiparent, szDirName, pFEAList));
    
        if (rc != NO_ERROR)
        {
            rc = ERROR_ACCESS_DENIED;
        }
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
    
    return rc;
}

int APIENTRY NdpFindStart (HCONNECTION conn, void *plist, NDFILEINFO *pfiparent, char *szPath, ULONG ulAttribute)
{
    LOG((ND_LL_DEBUG, "NdpFindStart: conn=%p, plist = %p, pfiparent=%p, path = %s, attr = %8.8X\r\n", conn, plist, pfiparent, szPath, ulAttribute));
    
    NdConnection *pConn = (NdConnection *)conn;
    
    (void)pfiparent;
    
    NDPATHELEMENT *pel = ph->fsphNameElem (0);
    
    char *pszParent = "";
    char *pszMask = pel->name;
        
    if (!ph->fsphIsLastElem (0, NULL))
    {
        /* this is not only element in the path */
            
        /* "aa\bb" -> "aa" */
        int l = strlen (szPath) - pel->length - 1;
            
        pszParent = (char *)alloca (l + 2);
                
        memcpy (pszParent, szPath, l);
               
        pszParent[l] = '\0';
    }
                    
    LOG((ND_LL_DEBUG, "NdpFindStart: pszParent = %s, pszMask = %s\r\n", pszParent, pszMask));
    
    NdPathInfo *ppathinfo = NULL;
                    
    int rc = NO_ERROR;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (pszParent);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = pConn->List (&ppathinfo, fullpath);
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
        
    if (rc == NO_ERROR)
    {
        while (ppathinfo)
        {
            LOG((ND_LL_DEBUG, "Matching: %s (%s)\r\n", ppathinfo->Name(), pszMask));
                
            if ( 
                ph->fsphAttrMatch (ulAttribute, ppathinfo->FileAttribute()) &&
                ph->fsphWildMatch (pszMask, ppathinfo->Name(), ND_IGNORE_CASE)
               )
            {
                FILESTATUS3L stat;
                ppathinfo->toFILESTATUS3L (&stat);
               
                LOG((ND_LL_DEBUG, "Adding: %s\r\n", ppathinfo->Name()));
            
                ph->fsphAddFile32L (plist, &stat, ppathinfo->Name (), ppathinfo->NameLength(), NULL, 0, 0);
            }
                
            NdPathInfo *next = ppathinfo->Next ();
            delete ppathinfo;
            ppathinfo = next;
        }
    }
    else
    {
        while (ppathinfo)
        {
            LOG((ND_LL_DEBUG, "NdpFindStart: deleting %p\r\n", ppathinfo));
            NdPathInfo *next = ppathinfo->Next ();
            delete ppathinfo;
            ppathinfo = next;
        }
    }
    
    LOG((ND_LL_DEBUG, "NdpFindStart: rc = %d, conn=%p, plist = %p, pfiparent=%p, path = %s, attr = %8.8X\r\n", rc, conn, plist, pfiparent, szPath, ulAttribute));
    
    return rc;
}


int APIENTRY NdpMove (HCONNECTION conn, NDFILEINFO *pfiDst, char *szDst, NDFILEINFO *pfiSrc, char *szSrc)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiDst;
    (void)pfiSrc;
    
    const char *src_fullpath = pConn->GetResource ()->QueryFullPath (szSrc);
    
    if (!src_fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        const char *dst_fullpath = pConn->GetResource ()->QueryFullPath (szDst);
    
        if (!dst_fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pConn->Move (dst_fullpath, src_fullpath);
    
            LOG((ND_LL_DEBUG, "NdpMove: rc = %d, conn=%p, pfiDst = %p, dst = %s, pfiSrc = %p, src = %s\r\n", rc, conn, pfiDst, szDst, pfiSrc, szSrc));
    
            pConn->GetResource ()->ReleaseFullPath (dst_fullpath);
        }
        
        pConn->GetResource ()->ReleaseFullPath (src_fullpath);
    }
    
    return rc;
}    

int APIENTRY NdpMove2 (HCONNECTION conn, HRESOURCE resDst, NDFILEINFO *pfiDst, char *szDst, NDFILEINFO *pfiSrc, char *szSrc)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiDst;
    (void)pfiSrc;
    
    const char *src_fullpath = pConn->GetResource ()->QueryFullPath (szSrc);
    
    if (!src_fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        const char *dst_fullpath = pConn->GetResource ()->QueryFullPath (szDst);
    
        if (!dst_fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pConn->MoveTo ((NdResource *)resDst, dst_fullpath, src_fullpath);
    
            LOG((ND_LL_DEBUG, "NdpMove: rc = %d, conn=%p, resDst = %p, pfiDst = %p, dst = %s, pfiSrc = %p, src = %s\r\n", rc, conn, resDst, pfiDst, szDst, pfiSrc, szSrc));
    
            pConn->GetResource ()->ReleaseFullPath (dst_fullpath);
        }
        
        pConn->GetResource ()->ReleaseFullPath (src_fullpath);
    }
    
    return rc;
}

int APIENTRY NdpChangeCase (HCONNECTION conn, char *szDst, NDFILEINFO *pfiSrc, char *szSrc, char *szNewName, ULONG ulNameLen)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiSrc;
    
    const char *src_fullpath = pConn->GetResource ()->QueryFullPath (szSrc);
    
    if (!src_fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        const char *dst_fullpath = pConn->GetResource ()->QueryFullPath (szDst);
    
        if (!dst_fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pConn->ChangeCase (dst_fullpath, src_fullpath, szNewName, ulNameLen);
    
            LOG((ND_LL_DEBUG, "NdpChangeCase: rc = %d, conn=%p, dst = %s, pfiSrc = %p, src = %s, newname = %s, namelen = %d\r\n", rc, conn, szDst, pfiSrc, szSrc, szNewName, ulNameLen));
    
            pConn->GetResource ()->ReleaseFullPath (dst_fullpath);
        }
        
        pConn->GetResource ()->ReleaseFullPath (src_fullpath);
    }
    
    return rc;
}

int APIENTRY NdpRename (HCONNECTION conn, char *szDst, NDFILEINFO *pfiSrc, char *szSrc, char *szNewName, ULONG ulNameLen)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiSrc;
    
    const char *src_fullpath = pConn->GetResource ()->QueryFullPath (szSrc);
    
    if (!src_fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        const char *dst_fullpath = pConn->GetResource ()->QueryFullPath (szDst);
    
        if (!dst_fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pConn->Rename (dst_fullpath, src_fullpath, szNewName, ulNameLen);
      
            LOG((ND_LL_DEBUG, "NdpRename: rc = %d, conn=%p, dst = %s, pfiSrc = %p, src = %s, newname = %s, namelen = %d\r\n", rc, conn, szDst, pfiSrc, szSrc, szNewName, ulNameLen));
    
            pConn->GetResource ()->ReleaseFullPath (dst_fullpath);
        }
        
        pConn->GetResource ()->ReleaseFullPath (src_fullpath);
    }
    
    return rc;
}

int APIENTRY NdpOpenExisting (HCONNECTION conn, NDFILEINFO *pfi, NDFILEHANDLE *phandle, char *szFileName, ULONG ulOpenMode, USHORT *pfNeedEA)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfi;
    
    NdFile *pfile = new NdFile (NdFile::fileOpen);
    
    if (!pfile)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        NDPATHELEMENT *pel = ph->fsphNameElem (0);
        
        const char *fullpath = pConn->GetResource()->QueryFullPath (szFileName);
     
        if (!fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pfile->Initialize (pel->name, pel->length, fullpath, ulOpenMode);
            
            pConn->GetResource()->ReleaseFullPath (fullpath);
        }
    
        if (rc == NO_ERROR)
        {
            rc = pConn->Open (pfile);
        
            if (rc == NO_ERROR)
            {
                *pfNeedEA = 0;
                *phandle = (NDFILEHANDLE)pfile;
            }
        }
        
        if (rc != NO_ERROR)
        {
            delete pfile;
        }
    }
    
    LOG((ND_LL_DEBUG, "NdpOpenExisting: rc = %d, conn = %p, pfi = %p, path = %s, pfile = %p\r\n", rc, conn, pfi, szFileName, pfile));
    
    return rc;
}


int APIENTRY NdpOpenReplaceL (HCONNECTION conn, NDFILEINFO *pfi, NDFILEHANDLE *phandle, char *szFileName, LONGLONG llSize, ULONG ulOpenMode, ULONG ulAttribute, FEALIST *pFEAList)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfi;
    (void)llSize;
    (void)ulAttribute;
    (void)pFEAList;
    
    NdFile *pfile = new NdFile (NdFile::fileReplace);
    
    if (!pfile)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        NDPATHELEMENT *pel = ph->fsphNameElem (0);
        
        const char *fullpath = pConn->GetResource()->QueryFullPath (szFileName);
     
        if (!fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pfile->Initialize (pel->name, pel->length, fullpath, ulOpenMode);
           
            pConn->GetResource()->ReleaseFullPath (fullpath);
        }
    
        if (rc == NO_ERROR)
        {
            rc = pConn->OpenReplace (pfile);
        
            if (rc == NO_ERROR)
            {
                *phandle = (NDFILEHANDLE)pfile;
            }
        }
        
        if (rc != NO_ERROR)
        {
            delete pfile;
        }
    }
    
    LOG((ND_LL_DEBUG, "NdpOpenReplaceL: rc = %d, conn = %p, pfi = %p, path = %s, pfile = %p\r\n", rc, conn, pfi, szFileName, pfile));
    
    return rc;
}

int APIENTRY NdpOpenCreateL (HCONNECTION conn, NDFILEINFO *pfiparent, NDFILEHANDLE *phandle, char *szFileName, LONGLONG llSize, ULONG ulOpenMode, ULONG ulAttribute, FEALIST *pFEAList)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfiparent;
    (void)llSize;
    (void)ulAttribute;
    (void)pFEAList;
    
    NdFile *pfile = new NdFile (NdFile::fileCreate);
    
    if (!pfile)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        NDPATHELEMENT *pel = ph->fsphNameElem (0);
        
        const char *fullpath = pConn->GetResource()->QueryFullPath (szFileName);
     
        if (!fullpath)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            rc = pfile->Initialize (pel->name, pel->length, fullpath, ulOpenMode);
           
            pConn->GetResource()->ReleaseFullPath (fullpath);
        }
    
        if (rc == NO_ERROR)
        {
            rc = pConn->OpenCreate (pfile);
        
            if (rc == NO_ERROR)
            {
                *phandle = (NDFILEHANDLE)pfile;
            }
        }
        
        if (rc != NO_ERROR)
        {
            delete pfile;
        }
    }
    
    LOG((ND_LL_DEBUG, "NdpOpenCreateL: rc = %d, conn = %p, pfiparent = %p, path = %s, pfile = %p\r\n", rc, conn, pfiparent, szFileName, pfile));
    
    return rc;
}

int APIENTRY NdpDeleteDir (HCONNECTION conn, NDFILEINFO *pfi, char *szDir)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    int rc = NO_ERROR;
    
    (void)pfi;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (szDir);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = pConn->DeleteDir (fullpath);
    
        LOG((ND_LL_DEBUG, "NdpDeleteDir: rc = %d, conn=%p: pfi = %p: path = %s\r\n", rc, conn, pfi, szDir));
        
        if (rc != NO_ERROR)
        {
            rc = ERROR_ACCESS_DENIED;
        }
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
    
    return rc;
}

int APIENTRY NdpSetFileAttribute (HCONNECTION conn, NDFILEINFO *pfi, char *szFileName, USHORT usAttr)
{
    NdConnection *pConn = (NdConnection *)conn;
    
    ULONG oldAttr = pfi->stat.attrFile;
    
    pfi->stat.attrFile = usAttr;
    
    NDPATHELEMENT *pel = ph->fsphNameElem (0);
    
    NdPathInfo pathinfo(&pfi->stat, pel->name, pel->length);
    
    int rc = NO_ERROR;
    
    const char *fullpath = pConn->GetResource()->QueryFullPath (szFileName);
     
    if (!fullpath)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        rc = pConn->SetPathInfo (fullpath, &pathinfo);
    
        if (rc != NO_ERROR)
        {
            rc = ERROR_ACCESS_DENIED;
        }
        
        pConn->GetResource()->ReleaseFullPath (fullpath);
    }
    
    if (rc != NO_ERROR)
    {
        pfi->stat.attrFile = oldAttr;
    }
    
    LOG((ND_LL_DEBUG, "NdpSetFileAttribute: rc = %d, conn=%p: pfi = %p: path = %s\r\n", rc, conn, pfi, szFileName));
    
    return rc;
}

int APIENTRY NdpFlush (HRESOURCE resource)
{
    (void)resource;
    return ERROR_NOT_SUPPORTED;
}

#pragma off(unreferenced)
int APIENTRY NdpIOCTL (int type, HRESOURCE resource, char *path, int function,
                       void *in, ULONG insize, PULONG poutlen)
#pragma on(unreferenced)
{
    (void)type;
    (void)resource;
    (void)path;
    (void)function;
    (void)in;
    (void)insize;
    (void)poutlen;
    return ERROR_NOT_SUPPORTED;
}
                                      

int APIENTRY NdpFileQueryInfo (HCONNECTION conn, NDFILEHANDLE handle, void *plist)
{
    NdConnection *pConn = (NdConnection *)conn;
    NdFile *file = (NdFile *)handle;
    
    FILESTATUS3L stat;
    
    int rc = ph->fsphQueryFileStatus (&stat);
    
    if (rc == NO_ERROR)
    {
        rc = ph->fsphAddFile32L (plist, &stat, file->Name (), file->NameLength (), NULL, 0, 0);
    }
    
    return rc;
}

int APIENTRY NdpFileEAQuery (HCONNECTION conn, NDFILEHANDLE handle, GEALIST *pGEAList, FEALIST *pFEAList)
{
    (void)conn;
    (void)handle;
    (void)pGEAList;
    (void)pFEAList;
    return ERROR_NOT_SUPPORTED;
}

int APIENTRY NdpFileEASet (HCONNECTION conn, NDFILEHANDLE handle, FEALIST *pFEAList)
{
    (void)conn;
    (void)handle;
    (void)pFEAList;
    return ERROR_NOT_SUPPORTED;
}

int APIENTRY NdpFileEASize (HCONNECTION conn, NDFILEHANDLE handle, ULONG *pulEASize)
{
    (void)conn;
    (void)handle;
    (void)pulEASize;
    return ERROR_NOT_SUPPORTED;
}

int APIENTRY NdpFileSetInfo (HCONNECTION conn, NDFILEHANDLE handle, NDFILEINFO *pfi)
{
    NdConnection *pConn = (NdConnection *)conn;
    NdFile *file = (NdFile *)handle;
    
    NdPathInfo pathinfo(&pfi->stat, file->Name (), file->NameLength ());
        
    return pConn->SetPathInfo (file->FullName (), &pathinfo);
}

int APIENTRY NdpFileSetFilePtrL (HCONNECTION conn, NDFILEHANDLE handle, LONGLONG llOffset, ULONG ulMethod, LONGLONG *pllActual)
{
    NdConnection *pConn = (NdConnection *)conn;
    NdFile *file = (NdFile *)handle;
    
    LONGLONG llFilePtr;
    LONGLONG llFileSize;
    LONGLONG llActual;
    
    ph->fsphQueryFilePtr (&llFilePtr);
    ph->fsphQueryFileSize (&llFileSize);
    
    int rc = ph->fsphSetFilePtr (llFileSize, llFilePtr, llOffset, ulMethod, &llActual);
    
    if (rc == NO_ERROR)
    {
        rc = file->SetFilePtr(pConn, llActual);
        
        if (rc == NO_ERROR)
        {
            *pllActual = llActual;
        }
    }
    
    LOG((ND_LL_DEBUG, "NdpFileSetFilePtrL: rc = %d, conn=%p: offset = %lld, method = %d\r\n", rc, conn, llOffset, ulMethod));
    
    return rc;
}


int APIENTRY NdpFileClose (HCONNECTION conn, NDFILEHANDLE handle)
{
    NdConnection *pConn = (NdConnection *)conn;
    NdFile *file = (NdFile *)handle;
    
    int rc = file->Close (pConn);
    
    delete file;
    
    LOG((ND_LL_DEBUG, "NdpFileClose: rc = %d, conn=%p, file = %p\r\n", rc, conn, file));
    
    return rc;
}


int APIENTRY NdpFileCommit (HCONNECTION conn, NDFILEHANDLE handle)
{
    (void)conn;
    (void)handle;
    return NO_ERROR;
}

int APIENTRY NdpFileNewSizeL (HCONNECTION conn, NDFILEHANDLE handle, LONGLONG llLen)
{
    NdConnection *pConn = (NdConnection *)conn;
    NdFile *file = (NdFile *)handle;
    
    int rc = file->NewSize (pConn, llLen);
    
    LOG((ND_LL_DEBUG, "NdpFileNewSizeL: rc = %d, conn = %p, len = %lld\r\n", rc, conn, llLen));
    
    return rc;
}

int APIENTRY NdpFileRead (HCONNECTION conn, NDFILEHANDLE handle, void *pBuffer, ULONG ulRead, ULONG *pulActual)
{
    NdConnection *pConn = (NdConnection *)conn;
    NdFile *file = (NdFile *)handle;
    
    int rc = file->Read (pConn, pBuffer, ulRead, pulActual);
    
    LOG((ND_LL_DEBUG, "NdpFileRead: rc = %d, conn=%p: read = %d\r\n", rc, conn, ulRead));
    
    return rc;
}

int APIENTRY NdpFileWrite (HCONNECTION conn, NDFILEHANDLE handle, void *pBuffer, ULONG ulWrite, ULONG *pulActual)
{
    NdConnection *pConn = (NdConnection *)conn;
    NdFile *file = (NdFile *)handle;

    int rc = file->Write (pConn, pBuffer, ulWrite, pulActual);
    
    LOG((ND_LL_DEBUG, "NdpFileWrite: rc = %d, conn=%p: write = %d\r\n", rc, conn, ulWrite));
    
    return rc;
}


NdPathInfo *NdPathInfo::Dublicate (NdPathInfo *ppathinfo)
{
    NdPathInfo *plist = new NdPathInfo (ppathinfo);
    
    if (plist)
    {
        NdPathInfo *iter = plist;
        ppathinfo = ppathinfo->next;
        
        while (ppathinfo)
        {
            NdPathInfo *p = new NdPathInfo (ppathinfo);
            
            if (!p)
            {
                DeleteList (plist);
                break;
            }
            
            iter->next = p;
            iter = p;
            
            ppathinfo = ppathinfo->next;
        }
    }
    
    return plist;
}

void NdPathInfo::DeleteList (NdPathInfo *plist)
{
    while (plist)
    {
        NdPathInfo *next = plist->Next ();
        delete plist;
        plist = next;
    }
}

NdPathInfo::NdPathInfo (FILESTATUS3L *pstat, char *n, int l)
    :
    next (NULL),
    namelen (l),
    ctime (0),
    mtime (0),
    size (0),
    type (File),
    hidden (0)
{
    memcpy (name, n, l);
    name[l] = 0;
    
    if (pstat->attrFile & FILE_DIRECTORY)
    {
        type = Dir;
    }
    
    size = pstat->cbFile;
    
    ph->fsphDosDateToUnixTime (pstat->fdateCreation, pstat->ftimeCreation, &ctime);
    ph->fsphDosDateToUnixTime (pstat->fdateLastWrite, pstat->ftimeLastWrite, &mtime);
}

NdPathInfo::NdPathInfo (void)
    :
    next (NULL),
    namelen (0),
    ctime (0),
    mtime (0),
    size (0),
    type (Unknown),
    hidden (0)
{
    name[0] = 0;
}

NdPathInfo::NdPathInfo (NdPathInfo *p)
    :
    next (NULL),
    namelen (p->namelen),
    ctime (p->ctime),
    mtime (p->mtime),
    size (p->size),
    type (p->type),
    hidden (p->hidden)
{
    memcpy (name, p->name, p->namelen + 1);
}

NdPathInfo &NdPathInfo::operator= (const NdPathInfo &r)
{
    namelen = r.namelen;
    ctime = r.ctime;
    mtime = r.mtime;
    size = r.size;
    type = r.type;
    hidden = r.hidden;
    memcpy (name, r.name, namelen);
    return *this;
}

void NdPathInfo::toFILESTATUS3L (FILESTATUS3L *pstat)
{
    ph->fsphUnixTimeToDosDate (ctime, &pstat->fdateCreation, &pstat->ftimeCreation);
    ph->fsphUnixTimeToDosDate (mtime, &pstat->fdateLastWrite, &pstat->ftimeLastWrite);
    pstat->fdateLastAccess = pstat->fdateLastWrite;
    pstat->ftimeLastAccess = pstat->ftimeLastWrite;
    pstat->attrFile = FileAttribute ();
    pstat->cbFile = size;
    pstat->cbFileAlloc = size;
}

ULONG NdPathInfo::FileAttribute (void)
{
    ULONG attr = hidden? (FILE_HIDDEN | FILE_SYSTEM): 0;
    
    if (type == File)
    {
        attr |= FILE_ARCHIVED;
    }
    if (type == Dir)
    {
        attr |= FILE_DIRECTORY;
    }
    
    return attr;
}
        
int NdResource::QueryStringProperty (const char *name, const char **ppvalue, unsigned long *plen)
{
    return ph->fsphQueryStringProperty (properties, name, ppvalue, plen);
}

int NdResource::QueryUlongProperty (const char *name, unsigned long *pvalue)
{
    return ph->fsphQueryUlongProperty (properties, name, pvalue);
}

int NdResource::SetProperty (const char *name, const char *pvalue)
{
    return ph->fsphSetProperty (properties, name, pvalue);
}

int NdResource::SetUlongProperty (const char *name, unsigned long value)
{
    return ph->fsphSetUlongProperty (properties, name, value);
}

    
NdFile::NdFile (int openflag)
    :
    ulNameLen (0),
    ulOpenMode (0),
    llFilePtr (0),
    iOpenFlag (openflag),
    iWritten (0)
{
    szFullName[0] = 0;
    szName[0] = 0;
}

inline NdFile::~NdFile (void)
{
}
        
int NdFile::Initialize (const char *name, int namelen, const char *fullname, ULONG openmode)
{
    int rc = NO_ERROR;
    
    memcpy (szName, name, namelen);
    szName[namelen] = 0;
    ulNameLen = namelen;
    
    strcpy (szFullName, fullname);
    
    ulOpenMode = openmode;
    
    return rc;
}

char *NdFile::FullName (void)
{
    return szFullName;
}

char *NdFile::Name (void)
{
    return szName;
}

int NdFile::NameLength (void)
{
    return ulNameLen;
}

int NdFile::Close(NdConnection *pConn)
{
    int rc = NO_ERROR;
    
    if (iOpenFlag == fileReplace ||
        iOpenFlag == fileCreate)
    {
        if (!iWritten)
        {
            ULONG ulActual;
            rc = SetFilePtr(pConn, 0);
            if (rc == NO_ERROR)
            {
                rc = Write (pConn, NULL, 0, &ulActual);
            }
        }
    }
    
    return rc;
}
    
int NdFile::NewSize (NdConnection *pConn, LONGLONG llLen)
{
    int rc = NO_ERROR;
    
    if ((ulOpenMode & 0x7) == OPEN_ACCESS_READONLY)
    {
        rc = ERROR_ACCESS_DENIED;
    }
    else
    {
        ULONG ulActual;
        rc = SetFilePtr(pConn, llLen);
        if (rc == NO_ERROR)
        {
            rc = Write (pConn, NULL, 0, &ulActual);
        }
    }
    
    return rc;
}

int NdFile::SetFilePtr (NdConnection *pconn, LONGLONG llOffset)
{
    (void)pconn;
    llFilePtr = llOffset;
    return NO_ERROR;
}

int NdFile::Read (NdConnection *pconn, void *pbuffer, ULONG ulRead, ULONG *pulActual)
{
    int rc = pconn->Read (szFullName, llFilePtr, pbuffer, ulRead, pulActual);
    
    if (rc == NO_ERROR)
    {
        llFilePtr += *pulActual;
    }
    
    return rc;
}

int NdFile::Write (NdConnection *pconn, void *pbuffer, ULONG ulWrite, ULONG *pulActual)
{
    int rc = NO_ERROR;
    
    if (iOpenFlag == fileReplace && !iWritten)
    {
        // if the file must be replaced we have to write 0 bytes to begin,
        // this will truncate the file
        if (llFilePtr != 0)
        {
            ULONG ulDummy;
            rc = pconn->Write (szFullName, llFilePtr, NULL, 0, &ulDummy);
            
            if (rc == NO_ERROR)
            {
                iWritten = 1;
            }
        }
    }
    
    if (rc == NO_ERROR)
    {
        rc = pconn->Write (szFullName, llFilePtr, pbuffer, ulWrite, pulActual);
        
        if (rc == NO_ERROR)
        {
            LOG((ND_LL_DEBUG, "advance file ptr (%lld) to %d\r\n", llFilePtr, *pulActual));
            llFilePtr += *pulActual;
        }
        
        iWritten = 1; // there was a write operation, even if it actually failed
    }
    
    return rc;
}
