/*  
 * Copyright (c) 2005 Blueprint Software Works
 *                    http://www.blueprintsoftwareworks.com
 * Copyright (c) 2005 Vitali E. Pelenyov
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
 * THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
 */

#ifndef __NDP__H
#define __NDP__H

#define INCL_LONGLONG
#include <ndfs.h>

#define NDPL_LARGEFILES
#include <ndextpl2.h>

#include <time.h>
#include <string.h>

class NdPathInfo
{
    private:
        NdPathInfo *next;
        
        char name[260];
        int namelen;
        
        NdPathInfo (NdPathInfo *p);
        
    public:
    
        static NdPathInfo *Dublicate (NdPathInfo *ppathinfo);
        static void DeleteList (NdPathInfo *plist);
    
        NdPathInfo (FILESTATUS3L *pstat, char *n, int l);
        NdPathInfo (void);
        virtual ~NdPathInfo (void) {};
        
        NdPathInfo &operator= (const NdPathInfo &r);
        
        NdPathInfo *Next (void) { return next; };
        void AppendNext (NdPathInfo *p) { next = p; };
        
        int isDir (void) { return type == Dir; };
        
        char *Name (void) {return name;};
        int NameLength (void) {return namelen;};
        
        void SetName (const char *n, int l) 
        {
            memcpy (name, n, l);
            name[l] = 0;
            namelen = l;
        };
        
        void toFILESTATUS3L (FILESTATUS3L *pstat);
        ULONG FileAttribute (void);
        
        time_t ctime;
        time_t mtime;
        long long size;
        unsigned type : 4;
        unsigned hidden : 1;
        
        enum
        {
            Unknown = 0,
            File = 1,
            Dir = 2
        };
};

class NdResource 
{
    private:
    
        NDPROPERTYHANDLE properties;
        
    public:
        NdResource (NDPROPERTYHANDLE p)
            :
            properties (p)
        {
        };
        
        virtual ~NdResource (void) 
        {
        };
        
        int QueryStringProperty (const char *name, const char **ppvalue, unsigned long *plen);
        int QueryUlongProperty (const char *name, unsigned long *pvalue);
        int SetProperty (const char *name, const char *pvalue);
        int SetUlongProperty (const char *name, unsigned long value);

        virtual int Initialize (void) = 0;
        virtual int EqualTo (NdResource *pres) = 0;
        virtual int Available (void) = 0;
        virtual int QueryInfoString (char **ppinfo) = 0;
        virtual void ReleaseInfoString (char *pinfo) = 0;

        virtual const char *QueryFullPath (const char *path) = 0;
        virtual void ReleaseFullPath (const char *fullpath) = 0;
};

class NdFile;

class NdConnection
{
    private:
        NdResource *presource;
     
    public:
        NdConnection (NdResource *pres)
            :
            presource (pres)
        {
        };
        
        virtual ~NdConnection (void) 
        {
        };
        
        NdResource *GetResource (void) {return presource;};
        
        virtual int Initialize (void) = 0;
        
        virtual int Refresh (const char *path, int tree) = 0;
    
        virtual int QueryPathInfo (NdPathInfo *ppathinfo, const char *path) = 0;
        
        virtual int SetPathInfo (const char *path, NdPathInfo *ppathinfo) = 0;
        virtual int SetCurrentDir (const char *path) = 0;
        
        virtual int Copy (const char *dst, const char *src, ULONG ulOption) = 0;
        virtual int CopyTo (NdResource *resdst, const char *dst, const char *src, ULONG ulOption) = 0;
        
        virtual int DeleteFile (const char *path) = 0;
        
        virtual int CreateDir (const char *path) = 0;
        
        virtual int List (NdPathInfo **pppathinfo, const char *path) = 0;
        
        virtual int Move (const char *dst, const char *src) = 0;
        virtual int MoveTo (NdResource *resdst, const char *dst, const char *src) = 0;

        virtual int ChangeCase (const char *dst, const char *src, const char *newname, ULONG namelen) = 0;
        virtual int Rename (const char *dst, const char *src, const char *newname, ULONG namelen) = 0;
    
        virtual int Open (NdFile *pfile) = 0;
        virtual int OpenReplace (NdFile *pfile) = 0;
        virtual int OpenCreate (NdFile *pfile) = 0;
        
        virtual int DeleteDir (const char *path) = 0;
    
        virtual int Read (const char *szFile, LONGLONG llOffset, void *pbuffer, ULONG ulRead, ULONG *pulActual) = 0;
        virtual int Write (const char *szFile, LONGLONG llOffset, void *pbuffer, ULONG ulWrite, ULONG *pulActual) = 0;
};

class NdFile
{
    private:
        char szFullName[260];
        char szName[260];
        ULONG ulNameLen;
        
        ULONG ulOpenMode;
        LONGLONG llFilePtr;
    
        int iOpenFlag;
        int iWritten;
        
    public:
        NdFile (int openflag);
        virtual ~NdFile (void);
        
        int Initialize (const char *name, int namelen, const char *fullname, ULONG openmode);
        
        char *FullName (void);
        char *Name (void);
        int NameLength (void);

        enum {
            fileOpen,
            fileCreate,
            fileReplace
        };
        
        int Close (NdConnection *pconn);
        int NewSize (NdConnection *pconn, LONGLONG llLen);
        int SetFilePtr (NdConnection *pconn, LONGLONG llOffset);
        int Read (NdConnection *pconn, void *pbuffer, ULONG ulRead, ULONG *pulActual);
        int Write (NdConnection *pconn, void *pbuffer, ULONG ulWrite, ULONG *pulActual);
};

#endif /* __NDP__H */
