/* 
   Basic HTTP and WebDAV methods, extension to neon library.
   Copyright (C) 1999-2004, Joe Orton <joe@manyfish.co.uk>
   Copyright (C) 2005, Vitali E. Pelenyov

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA

*/

#include "ne_nd.h"

#include "ne_i18n.h"

#include <string.h>

int ne_put_range_from_buffer(ne_session *sess, const char *uri, 
                             void *pbuffer, ne_content_range64 *range)
{
    ne_request *req = ne_request_create(sess, "PUT", uri);
    
    const ne_status *status;
    int ret;
    
    ne_print_request_header(req, "Content-Range", 
                            "bytes %" NE_FMT_OFF64_T "-%" NE_FMT_OFF64_T "/*",
                            range->start, range->end);
                            
    ne_set_request_body_buffer(req, (const char *)pbuffer, (size_t)(range->end - range->start + 1));

    ret = ne_request_dispatch(req);

    status = ne_get_status(req);

    if (status && status->code == 201) {
	ret = NE_OK;
    }
    else if (ret == NE_OK) {
	if (status && status->klass != 2) {
	    ret = NE_ERROR;
	}
    } 
    
    ne_request_destroy(req);
    
    return ret;
}

static int dispatch_to_buffer(ne_request *req, void *pbuffer, ne_content_range64 *range,
                              unsigned long *pactual)
{
    ne_session *const sess = ne_get_session(req);
    const ne_status *const st = ne_get_status(req);
    int ret;

    do {
        const char *value;
        
        ret = ne_begin_request(req);
        if (ret != NE_OK) break;

        value = ne_get_response_header(req, "Content-Range");

        /* For a 206 response, check that a Content-Range header is
         * given which matches the Range request header. */
        if (range && st->code == 206 
            && (value == NULL || strncmp(value, "bytes ", 6) != 0)) {
            ne_set_error(sess, _("Response did not include requested range"));
            return NE_ERROR;
        }

        if ((range && st->code == 206) || (!range && st->klass == 2)) {
            size_t buflen = (size_t)(range->end - range->start + 1);
            ret = ne_read_response_to_buffer(req, pbuffer, &buflen);
            *pactual = buflen;
        } else {
            ret = ne_discard_response(req);
        }

        if (ret == NE_OK) ret = ne_end_request(req);
    } while (ret == NE_RETRY);

    return ret;
}

int ne_get_range_to_buffer(ne_session *sess, const char *uri, 
                           void *pbuffer, ne_content_range64 *range,
                           unsigned long *pactual)
{
    ne_request *req = ne_request_create(sess, "GET", uri);
    const ne_status *status;
    int ret;
    char brange[64];

    if (range->end == -1) {
        ne_snprintf(brange, sizeof brange, "bytes=%" NE_FMT_OFF64_T "-", 
                    range->start);
    }
    else {
	ne_snprintf(brange, sizeof brange,
                    "bytes=%" NE_FMT_OFF64_T "-%" NE_FMT_OFF64_T,
                    range->start, range->end);
    }

    ne_add_request_header(req, "Range", brange);
    ne_add_request_header(req, "Accept-Ranges", "bytes");

    ret = dispatch_to_buffer(req, pbuffer, range, pactual);

    status = ne_get_status(req);

    if (ret == NE_OK && status->code == 416) {
	/* connection is terminated too early with Apache/1.3, so we check
	 * this even if ret == NE_ERROR... */
	ne_set_error(sess, _("Range is not satisfiable"));
	ret = NE_ERROR;
    }
    else if (ret == NE_OK) {
	if (status->klass == 2 && status->code != 206) {
	    ne_set_error(sess, _("Resource does not support ranged GETs."));
	    ret = NE_ERROR;
	}
	else if (status->klass != 2) {
	    ret = NE_ERROR;
	}
    } 
    
    ne_request_destroy(req);

    return ret;
}

