/*  
 * Copyright (c) 2005 Blueprint Software Works
 *                    http://www.blueprintsoftwareworks.com
 * Copyright (c) 2005 Vitali E. Pelenyov
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
 * THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
 */

#include "resource.hpp"

#include "debug.hpp"

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>

#include "ne_nd.h"

#if DEBUG
FILE *fdebug;
#endif

int ResourceInit (void)
{
    if (ne_sock_init()) 
    {
	LOG((ND_LL_FATAL, "Failed to initialize socket libraries.\r\n"));
	return ERROR_NETWORK_ACCESS_DENIED;
    }
    
#if DEBUG
    fdebug = fopen ("neon.log", "wb");
    ne_debug_init(fdebug, -1);
#endif

    return NO_ERROR;
}

int ResourceShutdown (void)
{
    // global termination
#if DEBUG
    fclose (fdebug);
#endif
    return NO_ERROR;
}

Resource::Resource (NDPROPERTYHANDLE p)
    :
    NdResource (p),
    fURIPathIsRoot (0)
{
}
        
/* virtual */ Resource::~Resource (void)
{
    ne_uri_free(&uri);
}
        
/* virtual */ int Resource::Initialize (void)
{
    // analyze resource properties here
    const char *str = NULL;
    ULONG len = 0;
    
    int rc = QueryStringProperty (NULL, &str, &len);
    
    if (rc != NO_ERROR || len == 0)
    {
        rc = QueryStringProperty ("URL", &str, &len);
    }
    
    if (rc == NO_ERROR)
    {
        if (len == 0 || ne_uri_parse(str, &uri) || uri.path == NULL || uri.host == NULL)
        {
            LOG((ND_LL_FATAL, "Invalid URI `%s'.\r\n", str));
            rc = ERROR_INVALID_PARAMETER;
        }
        else
        {
            if (uri.scheme == NULL)
            {
                uri.scheme = ne_strdup ("http");
            }
            if (uri.port == 0)
            {
                uri.port = ne_uri_defaultport(uri.scheme);
            }
            
LOG((ND_LL_DEBUG, "uri path: %s\r\n", uri.path));
            
            if (uri.path[0] == '/' && uri.path[1] == 0)
            {
LOG((ND_LL_DEBUG, "uri path: root\r\n"));
                fURIPathIsRoot = 1;
            }
            else
            {
                // remove last '/' if it exists
                char *lastslash = ph->fsphStrRChr (uri.path, '/');
                
                if (lastslash && lastslash[1] == 0)
                {
                    lastslash[0] = 0;
                }
            }
        }
    }
    
    if (rc == NO_ERROR)
    {
        char *pinfo = ne_uri_unparse (&uri);
    
        if (!pinfo)
        {
            rc = ERROR_NOT_ENOUGH_MEMORY;
        }
        else
        {
            SetProperty ("URL", pinfo);
            ne_free (pinfo);
        }
    }
    
    return rc;
}

/* virtual */ int Resource::EqualTo (NdResource *pres)
{
    Resource *pRes = (Resource *)pres;
    
    // compare this and pRes to see if they are equal,
    
    return (ne_uri_cmp (&uri, &pRes->uri) == 0);
}

/* virtual */ int Resource::Available (void)
{
    // check if resource physically available
    return 1;
}

/* virtual */ int Resource::QueryInfoString (char **ppinfo)
{
    char *pinfo = ne_uri_unparse (&uri);
    
    if (!pinfo)
    {
        return ERROR_NOT_ENOUGH_MEMORY;
    }
    
    *ppinfo = pinfo;
    return NO_ERROR;
}

/* virtual */ void Resource::ReleaseInfoString (char *pinfo)
{
    // free memory buffer returned in QueryInfoString (*ppinfo)
    ne_free (pinfo);
}

Session *Resource::CreateSession (void)
{
    ne_session *sess = ne_session_create(uri.scheme, uri.host, uri.port);
    
    if (!sess)
    {
        return NULL;
    }
    
    return new Session (sess);
}

void Resource::DeleteSession (Session *sess)
{
     delete sess;
}

const char *Resource::QueryURI (const char *fullpath)
{
    ne_buffer *buf = ne_buffer_create();

    ne_buffer_concat(buf, uri.scheme, "://", uri.host, NULL);

    if (uri.port > 0 && ne_uri_defaultport(uri.scheme) != uri.port) {
	char str[20];
	ne_snprintf(str, 20, ":%d", uri.port);
	ne_buffer_zappend(buf, str);
    }

    ne_buffer_zappend(buf, fullpath);

    return (const char *)ne_buffer_finish(buf);
}

void Resource::ReleaseURI (const char *res_uri)
{
    ne_free ((char *)res_uri);
}

const char *Resource::QueryFullPath (const char *path)
{
    // utf8 char can occupy up to 4 bytes
    int utf8uripathlen = 4*(strlen (uri.path) + 1);
    char *utf8uripath = (char *)alloca (utf8uripathlen);
    
    if (!ph->fsphCpTranslateString (ph->fsphQueryCurrentCp (), uri.path, 1208, utf8uripathlen, utf8uripath))
    {
        LOG((ND_LL_DEBUG, "Resource::QueryFullPath uri.path translation failed\r\n"));
        return NULL;
    }
    
    int utf8pathlen = 4*(strlen (path) + 1);
    char *utf8path = (char *)alloca (utf8pathlen);
    
    if (!ph->fsphCpTranslateString (ph->fsphQueryCurrentCp (), path, 1208, utf8pathlen, utf8path))
    {
        LOG((ND_LL_DEBUG, "Resource::QueryFullPath path translation failed\r\n"));
        return NULL;
    }
    
    const char *fullpath = dav_escape_uri (utf8uripath, utf8path);
    LOG((ND_LL_DEBUG, "QueryFullPath: %s + %s = %s\r\n", uri.path, path, fullpath));
    return fullpath;
}

void Resource::ReleaseFullPath (const char *fullpath)
{
    dav_free ((char *)fullpath);
}

Connection::Connection (Resource *pres)
    :
    NdConnection (pres),
    sess (NULL),
    cookie (NULL)
{
}
        
/* virtual */ Connection::~Connection (void)
{
    GetResource ()->DeleteSession (sess);
    
    if (cookie)
    {
        dav_free (cookie);
    }
}
        
int auth_creds (void *userdata, const char *realm, int attempt,
                char *username, char *password)
{
    Connection *pconn = (Connection *)userdata;
    
    (void)realm;
    
    if (attempt > 7)
    {
        return attempt;
    }
    
    strcpy (username, "");
    strcpy (password, "");
    
    if (pconn)
    {
        pconn->GetAuthInfo (username, password);
    }
    
    return 0;
}

void Connection::GetAuthInfo (char *username, char *password)
{
    const char *str = NULL;
    ULONG len = 0;
    
    int rc = GetResource()->QueryStringProperty ("USERNAME", &str, &len);
    
    if (rc == NO_ERROR && len < NE_ABUFSIZ)
    {
        strcpy (username, str);
        LOG((ND_LL_DEBUG, "Username %s\n", str));
    }
    
    rc = GetResource()->QueryStringProperty ("PASSWORD", &str, &len);
    
    if (rc == NO_ERROR && len < NE_ABUFSIZ)
    {
        strcpy (password, str);
        LOG((ND_LL_DEBUG, "Password length %d\n", len));
    }
}

int ssl_verify (void *userdata, int failures, const ne_ssl_certificate *cert)
{
    ne_session *sess = (ne_session *)userdata;
    
    (void)cert;
    
    LOG((ND_LL_DEBUG, "SSL failures %8.8X\n", failures));
    
#if 0
    ne_ssl_cert_write(cert, "newcert");
    ne_ssl_trust_cert(sess, cert);
#endif
    
    return 0;
}

void Connection::AddCookie (ne_request *req, ne_buffer *header)
{
    if (cookie)
    {
        ne_buffer_concat(header, "Cookie: ", cookie, "\r\n", NULL);
    }
    
    (void)req;
}

void pre_send (ne_request *req, void *userdata, ne_buffer *header)
{
    Connection *pconn = (Connection *)userdata;
    
    if (pconn)
    {
        pconn->AddCookie (req, header);
    }
}

void Connection::SetCookie (ne_request *req, const ne_status *status)
{
    const char *cookie_hdr = ne_get_response_header(req, "Set-Cookie");
   
    if (cookie_hdr)
    {
        if (cookie)
        {
            dav_free (cookie);
        }
        
        int l = strlen (cookie_hdr) + 1;
        cookie = (char *)dav_malloc (l);
        
        if (cookie)
        {
            memcpy (cookie, cookie_hdr, l);
            
            char *semicolon = strchr (cookie, ';');
            
            if (semicolon)
            {
                *semicolon = '\0';
            }
        }
    }
    
    (void)status;
    
    return;
}

int post_send (ne_request *req, void *userdata, const ne_status *status)
{
    Connection *pconn = (Connection *)userdata;
    
    if (pconn)
    {
        pconn->SetCookie (req, status);
    }
    
    return 0;
}

/* virtual */ int Connection::Initialize (void)
{
    int rc = NO_ERROR;
     
    sess = GetResource()->CreateSession ();
            
    if (!sess)
    {
        LOG((ND_LL_FATAL, "Could not create a session.\r\n"));
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        const char *str = NULL;
        ULONG len = 0;
    
        rc = GetResource()->QueryStringProperty ("USERNAME", &str, &len);
    
        if (rc == NO_ERROR && len > 0)
        {
            LOG((ND_LL_DEBUG, "Username specified %s\n", str));
            ne_set_server_auth(sess->s(), auth_creds, this);
            ne_ssl_set_verify(sess->s(), ssl_verify, sess->s());
            ne_hook_post_send(sess->s(), post_send, this);
            ne_hook_pre_send(sess->s(), pre_send, this);
            
#if 0
            ne_ssl_certificate *ca = ne_ssl_cert_read("C:\\NDFS\\newsert");
            if (ca == NULL)
            {
                LOG((ND_LL_FATAL, "Could not load cetificate.\r\n"));
            }
            ne_ssl_trust_cert(sess->s(), ca);
            ne_ssl_cert_free(ca);
#endif
        }
    }
     
    return rc;
}

/* virtual */ int Connection::Refresh (const char *path, int tree)
{
    (void)path;
    (void)tree;
    return NO_ERROR;
}

#define PROP_CREATIONDATE 0 
#define PROP_GETLASTMODIFIED 1
#define PROP_GETCONTENTLENGTH 2
#define PROP_RESOURCETYPE 3
#define PROP_ISCOLLECTION 4
#define PROP_ISHIDDEN 5

static ne_propname querypathinfo_props[] = {
    { "DAV:", "creationdate" },
    { "DAV:", "getlastmodified" },
    { "DAV:", "getcontentlength" },
    { "DAV:", "resourcetype" },
    { "DAV:", "iscollection" },
    { "DAV:", "ishidden" },
    { NULL }
};

// the querypathinfo_context structure holds information about
// current operation mode (QueryPathInfo or List).
typedef struct _querypathinfo_context
{
    int list;              // 0 - for QueryPathInfo,
                           // 1 - for List if resource is a directory
                           // 2 - for List if resource is a file
    int rc;                // an error code from callback
                           // this may be a NO_MEMORY error if
                           // a new NdPathInfo would fail
                           
    NdPathInfo *ppathinfo; // for QueryPathInfo this is a pointer
                           // to preallocated structure which must be filled
                           // for List this is NULL and is filled
                           // by the callback and is the head of list of
                           // structures which contain the directory
                           // file informations
                           
    const char *path;      // ignored for QueryPathInfo, for List
                           // it is the directory path being listed and
                           // used to skip the directory itself
//    int pathlen;           // length of the above path
} querypathinfo_context;

static int isblank (const char *value)
{
    while (*value && isspace (*value)) value++;
    return (*value == 0);
}
    
static void querypathinfo_callback (void *userdata, const char *res_uri,
                                    const ne_prop_result_set *results)
{
    querypathinfo_context *pctx = (querypathinfo_context *)userdata;

#if DEBUG
    int n;
    
    LOG((ND_LL_DEBUG, "Properties of %s:\n", res_uri));
    
    for (n = 0; querypathinfo_props[n].name != NULL; n++) 
    {
	const char *value = ne_propset_value(results, &querypathinfo_props[n]);
	
	if (value) {
	    LOG((ND_LL_DEBUG, " %s = %s\n", querypathinfo_props[n].name, value));
	} else {
	    const ne_status *st = ne_propset_status(results, &querypathinfo_props[n]);
	    LOG((ND_LL_DEBUG, " Failed for %s: %d %s\n", querypathinfo_props[n].name, 
		   st->code, st->reason_phrase));
	}
    }
#endif

    if (pctx->rc != NO_ERROR)
    {
        return;
    }

    LOG((ND_LL_DEBUG, "Resource uri: %s\r\n", res_uri));
    const char *remote_charset_href = dav_unescape_uri (res_uri);
    LOG((ND_LL_DEBUG, "Unescaped uri: %s\r\n", remote_charset_href));
    
    int remote_charset_href_len = strlen (remote_charset_href) + 1;
    
    const char *href = (const char *)alloca (remote_charset_href_len);
            
    if (!ph->fsphCpTranslateString (
            1208, remote_charset_href, 
            ph->fsphQueryCurrentCp (), remote_charset_href_len, (char *)href))
    {
        dav_free (remote_charset_href);
        return;
    }
    
    NdPathInfo *ppathinfo;
    ne_uri u;
             
    if (pctx->list)
    {
        if (ne_uri_parse(href, &u))
        {
            dav_free (remote_charset_href);
            return;
        }
    
        if (u.path == NULL) 
        {
            dav_free (remote_charset_href);
            ne_uri_free(&u);
            return;
        }

        LOG((ND_LL_DEBUG, "path is %s (ctx is %s)\r\n", u.path, pctx->path));
        
        // allocate a new path info
        ppathinfo = new NdPathInfo ();
        if (!ppathinfo)
        {
            pctx->rc = ERROR_NOT_ENOUGH_MEMORY;
            dav_free (remote_charset_href);
            ne_uri_free(&u);
            return;
        }
    }
    else
    {
        ppathinfo = pctx->ppathinfo;
    }
    
    // obtain properties
#define PROPVAL(a) ne_propset_value(results, &querypathinfo_props[a]);

    const char *value = PROPVAL(PROP_CREATIONDATE);
    
    if (value)
    {
        ppathinfo->ctime = ne_httpdate_parse(value);
    }
    
    value = PROPVAL(PROP_GETLASTMODIFIED);
    
    if (value)
    {
        ppathinfo->mtime = ne_httpdate_parse(value);
    }
    
    value = PROPVAL(PROP_GETCONTENTLENGTH);
    
    if (value)
    {
        ppathinfo->size = strtoll(value, NULL, 10);
    }
    
    // check if this a collection 
    value = PROPVAL(PROP_ISCOLLECTION);
    
    if (value && (atoi(value) == 1))
    {
        // server explicitely states that this is a collection
        ppathinfo->type = NdPathInfo::Dir;
    }
    else
    {
        value = PROPVAL(PROP_RESOURCETYPE);
        
        if (value)
        {   
            if (strstr (value, "collection"))
            {
                ppathinfo->type = NdPathInfo::Dir;
            }
            else if (isblank (value))
            {
                ppathinfo->type = NdPathInfo::File;
            }
        }
    }
    
    // check if this resource is hidden 
    value = PROPVAL(PROP_ISHIDDEN);
    
    if (value)
    {
        ppathinfo->hidden = (atoi (value) != 0);
    }
    
    if (pctx->list)
    {
        int listfile = 0;
        
        if (ne_path_compare (pctx->path, res_uri) == 0)
        {
            // This is the resource being listed, if it is a file then
            // actually the file was mounted as resource and we have to 
            // list it, otherwise it is a regular directory listing
            // and we have to skip the directory.
            
            if (ppathinfo->type != NdPathInfo::File)
            {
                LOG((ND_LL_DEBUG, "Skipping directory.\n"));
                delete ppathinfo;
                dav_free (remote_charset_href);
                ne_uri_free(&u);
                return;
            }
            else
            {
                listfile = 1;
            }
        }
        
        // extract and save name of the resource
        
        char *lastslash = ph->fsphStrRChr (u.path, '/');
        if (lastslash && lastslash[1] == 0)
        {
            *lastslash = 0; // remove trailing slash
        }
        
        lastslash = ph->fsphStrRChr (u.path, '/');
        if (lastslash)
        {
            lastslash++;
            
            ppathinfo->SetName (lastslash, strlen (lastslash));
        }
        
        if (ppathinfo->NameLength () == 0)
        {
            // somehow there is no name for the uri, skip
            delete ppathinfo;
        }
        else
        {
            ppathinfo->AppendNext (pctx->ppathinfo);
            pctx->ppathinfo = ppathinfo;
        }
        
        ne_uri_free(&u);
    }
    
    dav_free (remote_charset_href);
    
    return;
#undef PROPVAL
}

/* virtual */ int Connection::QueryPathInfo (NdPathInfo *ppathinfo, const char *path)
{
    int rc = NO_ERROR;
    
    LOG((ND_LL_DEBUG, "Connection::QueryPathInfo: path = %s\r\n", path));
    
    querypathinfo_context ctx = 
    {
        0,
        NO_ERROR,
        ppathinfo,
        NULL
    };
       
    ppathinfo->type = NdPathInfo::Unknown;
    
    if (ne_simple_propfind(sess->s(), path, NE_DEPTH_ZERO, querypathinfo_props,
                           querypathinfo_callback, &ctx)) 
    {
        LOG((ND_LL_DEBUG, "npropget: PROPFIND failed: %s\n", ne_get_error(sess->s())));
        rc = ERROR_PATH_NOT_FOUND;
    }
    else
    {
        if (ppathinfo->type == NdPathInfo::Unknown)
        {
            rc = ERROR_PATH_NOT_FOUND;
        }
        else
        {
            // todo set real remote file name in the callback, right name case
            NDPATHELEMENT *pel = ph->fsphNameElem (0);
            ppathinfo->SetName (pel->name, pel->length);
        }
    }
    
    return rc;
}

/* virtual */ int Connection::SetPathInfo (const char *path, NdPathInfo *ppathinfo)
{
    (void)path;
    (void)ppathinfo;
    return NO_ERROR;
}

/* virtual */ int Connection::SetCurrentDir (const char *path)
{
    (void)path;
    return NO_ERROR;
}

/* virtual */ int Connection::Copy (const char *dst, const char *src, ULONG ulOption)
{
    int rc = NO_ERROR;
    
    if (ulOption & DCPY_APPEND)
    {
        return ERROR_NOT_SUPPORTED;
    }
    
    int overwrite = ((ulOption & DCPY_EXISTING) != 0);
    
    if (ne_copy(sess->s(), overwrite, NE_DEPTH_INFINITE, src, dst))
    {
        rc = ERROR_ACCESS_DENIED;
    }
    
    return rc;
}

/* virtual */ int Connection::CopyTo (NdResource *resdst, const char *dst, const char *src, ULONG ulOption)
{
    int rc = NO_ERROR;
    
    (void)resdst;
    
    if (ulOption & DCPY_APPEND)
    {
        return ERROR_NOT_SUPPORTED;
    }
    
    int overwrite = ((ulOption & DCPY_EXISTING) != 0);
    
    if (ne_copy(sess->s(), overwrite, NE_DEPTH_INFINITE, src, dst))
    {
        rc = ERROR_ACCESS_DENIED;
    }
    
    return rc;
}

        
/* virtual */ int Connection::DeleteFile (const char *path)
{
    int rc = NO_ERROR;
    
    if (ne_delete(sess->s(), path))
    {
        rc = ERROR_ACCESS_DENIED;
    }
    
    return rc;
}

        
/* virtual */ int Connection::CreateDir (const char *path)
{
    int rc = NO_ERROR;
    
    if (ne_mkcol(sess->s(), path))
    {
        rc = ERROR_ACCESS_DENIED;
    }
    
    return rc;
}

        
/* virtual */ int Connection::List (NdPathInfo **pppathinfo, const char *path)
{
    int rc = NO_ERROR;
    
    querypathinfo_context ctx = 
    {
        1,
        NO_ERROR,
        NULL,
        GetResource ()->QueryURI (path)
    };
        
    if (!ctx.path)
    {
        rc = ERROR_NOT_ENOUGH_MEMORY;
    }
    else
    {
        if (ne_simple_propfind(sess->s(), path, NE_DEPTH_ONE, querypathinfo_props,
                               querypathinfo_callback, &ctx)) 
        {
            printf("npropget: PROPFIND failed: %s\n", ne_get_error(sess->s()));
            rc = ERROR_NO_MORE_FILES;
        }
        else
        {
            if (ctx.rc != NO_ERROR)
            {
                rc = ctx.rc;
            }
            else
            {
                *pppathinfo = ctx.ppathinfo;
            }
        }
    }
    
    if (ctx.path)
    {
        GetResource ()->ReleaseURI (ctx.path);
    }
        
    return rc;
}

       
/* virtual */ int Connection::Move (const char *dst, const char *src)
{
    int rc = NO_ERROR;
    
    if (ne_move(sess->s(), 1, src, dst))
    {
        rc = ERROR_ACCESS_DENIED;
    }
    
    return rc;
}

/* virtual */ int Connection::MoveTo (NdResource *resdst, const char *dst, const char *src)
{
    int rc = NO_ERROR;
    
    (void)resdst;
    
    if (ne_move(sess->s(), 1, src, dst))
    {
        rc = ERROR_ACCESS_DENIED;
    }
    
    return rc;
}


/* virtual */ int Connection::ChangeCase (const char *dst, const char *src, const char *newname, ULONG namelen)
{
    (void)newname;
    (void)namelen;
    return Move (dst, src);
}

/* virtual */ int Connection::Rename (const char *dst, const char *src, const char *newname, ULONG namelen)
{
    (void)newname;
    (void)namelen;
    return Move (dst, src);
}

    
/* virtual */ int Connection::Open (NdFile *pfile)
{
    // nothing to do for a regular open
    (void)pfile;
    return NO_ERROR;
}

/* virtual */ int Connection::OpenReplace (NdFile *pfile)
{
    // todo just mark that first write must be not ranged
    DeleteFile (pfile->FullName ());
    return NO_ERROR;
}

/* virtual */ int Connection::OpenCreate (NdFile *pfile)
{
    // File will be actually created on a first write or on close
    (void)pfile;
    return NO_ERROR;
}

        
/* virtual */ int Connection::DeleteDir (const char *path)
{
    int rc = NO_ERROR;
    
    if (ne_delete(sess->s(), path))
    {
        rc = ERROR_ACCESS_DENIED;
    }
    
    return rc;
}

/* virtual */ int Connection::Read (const char *szFile, LONGLONG llOffset, void *pbuffer, ULONG ulRead, ULONG *pulActual)
{
    int rc = NO_ERROR;
    
    *pulActual = 0;

    if (ulRead > 0)
    {
        ne_content_range64 range = { llOffset, llOffset + ulRead - 1, ulRead };
    
        ne_get_range_to_buffer(sess->s(), szFile, pbuffer, &range, pulActual);
    }
    
    return rc;
}

/* virtual */ int Connection::Write (const char *szFile, LONGLONG llOffset, void *pbuffer, ULONG ulWrite, ULONG *pulActual)
{
    int rc = NO_ERROR;
    
    *pulActual = 0;

    if (ulWrite > 0)
    {
        ne_content_range64 range = { llOffset, llOffset + ulWrite - 1, ulWrite };
    
        if (ne_put_range_from_buffer(sess->s(), szFile, pbuffer, &range))
        {
            rc = ERROR_ACCESS_DENIED;
        }
        else
        {
            *pulActual = ulWrite;
        }
    }
    
    return rc;
}

